import nose

from chainer.testing import array
from chainer.testing import helper
from chainer.testing import parameterized
from chainer.testing import unary_math_function_test


assert_allclose = array.assert_allclose

parameterize = parameterized.parameterize
product = parameterized.product
product_dict = parameterized.product_dict

unary_math_function_unittest = \
    unary_math_function_test.unary_math_function_unittest

with_requires = helper.with_requires


def run_module(name, file):
    """Run current test cases of the file.

    Args:
        name: __name__ attribute of the file.
        file: __file__ attribute of the file.
    """

    if name == '__main__':

        nose.runmodule(argv=[file, '-vvs', '-x', '--pdb', '--pdb-failure'],
                       exit=False)



from chainer.iterators import multiprocess_iterator
from chainer.iterators import serial_iterator


MultiprocessIterator = multiprocess_iterator.MultiprocessIterator
SerialIterator = serial_iterator.SerialIterator



from chainer import cuda
from chainer import function
from chainer import utils
from chainer.utils import type_check


class Floor(function.Function):

    @property
    def label(self):
        return 'floor'

    def check_type_forward(self, in_types):
        type_check.expect(
            in_types.size() == 1,
            in_types[0].dtype.kind == 'f',
        )

    def forward(self, x):
        xp = cuda.get_array_module(*x)
        return utils.force_array(xp.floor(x[0]), x[0].dtype),

    def backward(self, x, grad_outputs):
        xp = cuda.get_array_module(*x)
        return xp.zeros_like(x[0]),


def floor(x):
    """Elementwise floor function.

    .. math::
       y_i = \\lfloor x_i \\rfloor

    Args:
        x (~chainer.Variable): Input variable.

    Returns:
        ~chainer.Variable: Output variable.
    """
    return Floor()(x)



from chainer.dataset import convert
from chainer.dataset import dataset_mixin
from chainer.dataset import download
from chainer.dataset import iterator


DatasetMixin = dataset_mixin.DatasetMixin
Iterator = iterator.Iterator

concat_examples = convert.concat_examples

get_dataset_root = download.get_dataset_root
set_dataset_root = download.set_dataset_root
get_dataset_directory = download.get_dataset_directory
cached_download = download.cached_download
cache_or_load_file = download.cache_or_load_file



from __future__ import print_function
import copy
import distutils
import os
from os import path
import shutil
import subprocess
import sys
import tempfile

import pkg_resources
import setuptools
from setuptools.command import build_ext

from install import build
from install import utils


dummy_extension = setuptools.Extension('chainer', ['chainer.c'])
cython_version = '0.23.0'
MODULES = [
    {
        'name': 'cuda',
        'file': [
            'cupy.core.core',
            'cupy.core.flags',
            'cupy.core.internal',
            'cupy.cuda.cublas',
            'cupy.cuda.curand',
            'cupy.cuda.device',
            'cupy.cuda.driver',
            'cupy.cuda.memory',
            'cupy.cuda.pinned_memory',
            'cupy.cuda.profiler',
            'cupy.cuda.nvtx',
            'cupy.cuda.function',
            'cupy.cuda.runtime',
            'cupy.util',
        ],
        'include': [
            'cublas_v2.h',
            'cuda.h',
            'cuda_profiler_api.h',
            'cuda_runtime.h',
            'curand.h',
            'nvToolsExt.h',
        ],
        'libraries': [
            'cublas',
            'cuda',
            'cudart',
            'curand',
            'nvToolsExt',
        ],
        'check_method': build.check_cuda_version,
    },
    {
        'name': 'cudnn',
        'file': [
            'cupy.cuda.cudnn',
        ],
        'include': [
            'cudnn.h',
        ],
        'libraries': [
            'cudnn',
        ],
        'check_method': build.check_cudnn_version,
    }
]

if sys.platform == 'win32':
    mod_cuda = MODULES[0]
    mod_cuda['file'].remove('cupy.cuda.nvtx')
    mod_cuda['include'].remove('nvToolsExt.h')
    mod_cuda['libraries'].remove('nvToolsExt')


def localpath(*args):
    return path.abspath(path.join(path.dirname(__file__), *args))


def check_include(dirs, file_path):
    return any(path.exists(path.join(dir, file_path)) for dir in dirs)


def check_readthedocs_environment():
    return os.environ.get('READTHEDOCS', None) == 'True'


def check_library(compiler, includes=(), libraries=(),
                  include_dirs=(), library_dirs=()):
    temp_dir = tempfile.mkdtemp()

    try:
        source = '''
        int main(int argc, char* argv[]) {
          return 0;
        }
        '''
        fname = os.path.join(temp_dir, 'a.cpp')
        with open(fname, 'w') as f:
            for header in includes:
                f.write('#include <%s>\n' % header)
            f.write(source)

        try:
            objects = compiler.compile([fname], output_dir=temp_dir,
                                       include_dirs=include_dirs)
        except distutils.errors.CompileError:
            return False

        try:
            compiler.link_shared_lib(objects,
                                     os.path.join(temp_dir, 'a'),
                                     libraries=libraries,
                                     library_dirs=library_dirs,
                                     target_lang='c++')
        except (distutils.errors.LinkError, TypeError):
            return False

        return True

    except distutils.errors.DistutilsError as e:
        print('distutils raises an error: %s' % e)
        return False

    finally:
        shutil.rmtree(temp_dir, ignore_errors=True)


def make_extensions(options, compiler):
    """Produce a list of Extension instances which passed to cythonize()."""

    no_cuda = options['no_cuda']
    settings = build.get_compiler_setting()

    include_dirs = settings['include_dirs']

    settings['include_dirs'] = [
        x for x in include_dirs if path.exists(x)]
    settings['library_dirs'] = [
        x for x in settings['library_dirs'] if path.exists(x)]
    if sys.platform != 'win32':
        settings['runtime_library_dirs'] = settings['library_dirs']
    if sys.platform == 'darwin':
        args = settings.setdefault('extra_link_args', [])
        args.append(
            '-Wl,' + ','.join('-rpath,' + p
                              for p in settings['library_dirs']))
        # -rpath is only supported when targetting Mac OS X 10.5 or later
        args.append('-mmacosx-version-min=10.5')

    # This is a workaround for Anaconda.
    # Anaconda installs libstdc++ from GCC 4.8 and it is not compatible
    # with GCC 5's new ABI.
    settings['define_macros'].append(('_GLIBCXX_USE_CXX11_ABI', '0'))

    if options['linetrace']:
        settings['define_macros'].append(('CYTHON_TRACE', '1'))
        settings['define_macros'].append(('CYTHON_TRACE_NOGIL', '1'))
    if no_cuda:
        settings['define_macros'].append(('CUPY_NO_CUDA', '1'))

    ret = []
    for module in MODULES:
        print('Include directories:', settings['include_dirs'])
        print('Library directories:', settings['library_dirs'])

        if not no_cuda:
            if not check_library(compiler,
                                 includes=module['include'],
                                 include_dirs=settings['include_dirs']):
                utils.print_warning(
                    'Include files not found: %s' % module['include'],
                    'Skip installing %s support' % module['name'],
                    'Check your CFLAGS environment variable')
                continue

            if not check_library(compiler,
                                 libraries=module['libraries'],
                                 library_dirs=settings['library_dirs']):
                utils.print_warning(
                    'Cannot link libraries: %s' % module['libraries'],
                    'Skip installing %s support' % module['name'],
                    'Check your LDFLAGS environment variable')
                continue

            if 'check_method' in module and \
               not module['check_method'](compiler, settings):
                continue

        s = settings.copy()
        if not no_cuda:
            s['libraries'] = module['libraries']
        ret.extend([
            setuptools.Extension(f, [path.join(*f.split('.')) + '.pyx'], **s)
            for f in module['file']])
    return ret


_arg_options = {}


def parse_args():
    global _arg_options
    _arg_options['profile'] = '--cupy-profile' in sys.argv
    if _arg_options['profile']:
        sys.argv.remove('--cupy-profile')

    cupy_coverage = '--cupy-coverage' in sys.argv
    if cupy_coverage:
        sys.argv.remove('--cupy-coverage')
    _arg_options['linetrace'] = cupy_coverage
    _arg_options['annotate'] = cupy_coverage

    _arg_options['no_cuda'] = '--cupy-no-cuda' in sys.argv
    if _arg_options['no_cuda']:
        sys.argv.remove('--cupy-no-cuda')
    if check_readthedocs_environment():
        _arg_options['no_cuda'] = True


def get_cython_pkg():
    try:
        return pkg_resources.get_distribution('cython')
    except pkg_resources.DistributionNotFound:
        return None


def run_command(cmd):
    try:
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        msg = '''Command %r failed:

  command: %s
  return code: %d
  output:

%s''' % (cmd[0], ' '.join(e.cmd), e.returncode, e.output)
        raise distutils.errors.DistutilsExecError(msg)


def cythonize(
        extensions, force=False, annotate=False, compiler_directives=None):
    cython_location = get_cython_pkg().location
    cython_path = path.join(cython_location, 'cython.py')
    print("cython path:%s" % cython_location)
    cmd = [sys.executable, cython_path]
    run_command(cmd + ['--version'])

    cmd.extend(['--fast-fail', '--verbose', '--cplus'])
    if compiler_directives is not None:
        for i in compiler_directives.items():
            cmd.append('--directive')
            cmd.append('%s=%s' % i)

    for ext in extensions:
        run_command(cmd + ext.sources)


def to_cpp_extensions(extensions):
    ret = []
    for x in extensions:
        ext = copy.copy(x)
        ext.sources = [path.splitext(f)[0] + ".cpp" for f in x.sources]
        ret.append(ext)
    return ret


def check_extensions(extensions):
    for x in extensions:
        for f in x.sources:
            if not path.isfile(f):
                msg = ('Missing file: %s\n' % f +
                       'Please install Cython.\n' +
                       'See http://docs.chainer.org/en/stable/install.html')
                raise RuntimeError(msg)


class chainer_build_ext(build_ext.build_ext):

    """`build_ext` command for cython files."""

    def finalize_options(self):
        ext_modules = self.distribution.ext_modules
        if dummy_extension in ext_modules:
            print('Executing cythonize')
            print('Options:', _arg_options)

            directive_keys = ('linetrace', 'profile')
            directives = {key: _arg_options[key] for key in directive_keys}

            cythonize_option_keys = ('annotate',)
            cythonize_options = {
                key: _arg_options[key] for key in cythonize_option_keys}

            compiler = distutils.ccompiler.new_compiler(compiler=self.compiler)
            distutils.sysconfig.customize_compiler(compiler)

            extensions = make_extensions(_arg_options, compiler)

            cython = get_cython_pkg()
            req_version = pkg_resources.parse_version(cython_version)
            if cython is not None and cython.parsed_version > req_version:
                cythonize(extensions, force=True,
                          compiler_directives=directives, **cythonize_options)

            extensions = to_cpp_extensions(extensions)
            check_extensions(extensions)

            # Modify ext_modules for cython
            ext_modules.remove(dummy_extension)
            ext_modules.extend(extensions)

        build_ext.build_ext.finalize_options(self)



import numpy

from chainer import cuda
from chainer import function
from chainer import utils
from chainer.utils import type_check


class Log1p(function.Function):

    @property
    def label(self):
        return 'log1p'

    def check_type_forward(self, in_types):
        type_check.expect(in_types.size() == 1)
        type_check.expect(in_types[0].dtype.kind == 'f')

    def forward_cpu(self, x):
        return utils.force_array(numpy.log1p(x[0])),

    def forward_gpu(self, x):
        return cuda.cupy.log1p(x[0]),

    def backward(self, x, gy):
        return utils.force_array(gy[0] / (x[0] + x[0].dtype.type(1.0))),


def log1p(x):
    """Elementwise natural logarithm plus one function."""
    return Log1p()(x)



from chainer.training.triggers import interval
from chainer.training.triggers import minmax_value_trigger


IntervalTrigger = interval.IntervalTrigger
MaxValueTrigger = minmax_value_trigger.MaxValueTrigger
MinValueTrigger = minmax_value_trigger.MinValueTrigger



"""Collection of :class:`~chainer.Function` implementations."""

from chainer.functions.activation import clipped_relu
from chainer.functions.activation import crelu
from chainer.functions.activation import elu
from chainer.functions.activation import hard_sigmoid
from chainer.functions.activation import leaky_relu
from chainer.functions.activation import log_softmax
from chainer.functions.activation import lstm
from chainer.functions.activation import maxout
from chainer.functions.activation import prelu
from chainer.functions.activation import relu
from chainer.functions.activation import sigmoid
from chainer.functions.activation import slstm
from chainer.functions.activation import softmax
from chainer.functions.activation import softplus
from chainer.functions.activation import tanh
from chainer.functions.array import broadcast
from chainer.functions.array import cast
from chainer.functions.array import concat
from chainer.functions.array import copy
from chainer.functions.array import depth2space
from chainer.functions.array import dstack
from chainer.functions.array import expand_dims
from chainer.functions.array import flatten
from chainer.functions.array import get_item
from chainer.functions.array import hstack
from chainer.functions.array import permutate
from chainer.functions.array import reshape
from chainer.functions.array import rollaxis
from chainer.functions.array import select_item
from chainer.functions.array import separate
from chainer.functions.array import space2depth
from chainer.functions.array import split_axis
from chainer.functions.array import squeeze
from chainer.functions.array import stack
from chainer.functions.array import swapaxes
from chainer.functions.array import tile
from chainer.functions.array import transpose
from chainer.functions.array import transpose_sequence
from chainer.functions.array import vstack
from chainer.functions.array import where
from chainer.functions.connection import bilinear
from chainer.functions.connection import convolution_2d
from chainer.functions.connection import convolution_nd
from chainer.functions.connection import deconvolution_2d
from chainer.functions.connection import deconvolution_nd
from chainer.functions.connection import dilated_convolution_2d
from chainer.functions.connection import embed_id
from chainer.functions.connection import linear
from chainer.functions.connection import n_step_lstm
from chainer.functions.evaluation import accuracy
from chainer.functions.evaluation import binary_accuracy
from chainer.functions.evaluation import classification_summary \
    as classification_summary_
from chainer.functions.loss import black_out
from chainer.functions.loss import contrastive
from chainer.functions.loss import crf1d
from chainer.functions.loss import cross_covariance
from chainer.functions.loss import ctc
from chainer.functions.loss import hinge
from chainer.functions.loss import huber_loss
from chainer.functions.loss import mean_squared_error
from chainer.functions.loss import negative_sampling
from chainer.functions.loss import sigmoid_cross_entropy
from chainer.functions.loss import softmax_cross_entropy
from chainer.functions.loss import triplet
from chainer.functions.loss import vae  # NOQA
from chainer.functions.math import basic_math
from chainer.functions.math import batch_l2_norm_squared
from chainer.functions.math import bias
from chainer.functions.math import ceil
from chainer.functions.math import clip
from chainer.functions.math import det
from chainer.functions.math import exponential
from chainer.functions.math import exponential_m1
from chainer.functions.math import floor
from chainer.functions.math import hyperbolic
from chainer.functions.math import identity
from chainer.functions.math import inv
from chainer.functions.math import linear_interpolate
from chainer.functions.math import logarithm_1p
from chainer.functions.math import logsumexp
from chainer.functions.math import matmul
from chainer.functions.math import maximum
from chainer.functions.math import minimum
from chainer.functions.math import minmax
from chainer.functions.math import scale
from chainer.functions.math import sqrt
from chainer.functions.math import square
from chainer.functions.math import squared_difference
from chainer.functions.math import sum
from chainer.functions.math import trigonometric
from chainer.functions.noise import dropout
from chainer.functions.noise import gaussian
from chainer.functions.noise import zoneout
from chainer.functions.normalization import batch_normalization
from chainer.functions.normalization import l2_normalization
from chainer.functions.normalization import local_response_normalization
from chainer.functions.pooling import average_pooling_2d
from chainer.functions.pooling import max_pooling_2d
from chainer.functions.pooling import roi_pooling_2d
from chainer.functions.pooling import spatial_pyramid_pooling_2d
from chainer.functions.pooling import unpooling_2d
from chainer.functions.util import forget
from chainer.links.activation import prelu as links_prelu
from chainer.links.connection import bilinear as links_bilinear
from chainer.links.connection import convolution_2d as links_convolution_2d
from chainer.links.connection import dilated_convolution_2d \
    as links_dilated_convolution_2d
from chainer.links.connection import embed_id as links_embed_id
from chainer.links.connection import inception
from chainer.links.connection import inceptionbn
from chainer.links.connection import linear as links_linear
from chainer.links.connection import parameter
from chainer.links.loss import hierarchical_softmax
from chainer.links.loss import negative_sampling as links_negative_sampling
from chainer.links.normalization import batch_normalization \
    as links_batch_normalization


ClippedReLU = clipped_relu.ClippedReLU
clipped_relu = clipped_relu.clipped_relu
CReLU = crelu.CReLU
crelu = crelu.crelu
ConnectionistTemporalClassification = ctc.ConnectionistTemporalClassification
connectionist_temporal_classification \
    = ctc.connectionist_temporal_classification
ELU = elu.ELU
elu = elu.elu
HardSigmoid = hard_sigmoid.HardSigmoid
hard_sigmoid = hard_sigmoid.hard_sigmoid
LeakyReLU = leaky_relu.LeakyReLU
leaky_relu = leaky_relu.leaky_relu
LogSoftmax = log_softmax.LogSoftmax
log_softmax = log_softmax.log_softmax
LSTM = lstm.LSTM
lstm = lstm.lstm
maxout = maxout.maxout
prelu = prelu.prelu
ReLU = relu.ReLU
relu = relu.relu
Sigmoid = sigmoid.Sigmoid
sigmoid = sigmoid.sigmoid
SLSTM = slstm.SLSTM
slstm = slstm.slstm
Softmax = softmax.Softmax
softmax = softmax.softmax
Softplus = softplus.Softplus
softplus = softplus.softplus
Tanh = tanh.Tanh
tanh = tanh.tanh


Broadcast = broadcast.Broadcast
BroadcastTo = broadcast.BroadcastTo
broadcast_to = broadcast.broadcast_to
broadcast = broadcast.broadcast
Cast = cast.Cast
cast = cast.cast
Concat = concat.Concat
concat = concat.concat
Copy = copy.Copy
copy = copy.copy
Depth2Space = depth2space.Depth2Space
depth2space = depth2space.depth2space
dstack = dstack.dstack
ExpandDims = expand_dims.ExpandDims
expand_dims = expand_dims.expand_dims
Flatten = flatten.Flatten
flatten = flatten.flatten
GetItem = get_item.GetItem
get_item = get_item.get_item
hstack = hstack.hstack
Permutate = permutate.Permutate
permutate = permutate.permutate
Reshape = reshape.Reshape
reshape = reshape.reshape
Rollaxis = rollaxis.Rollaxis
rollaxis = rollaxis.rollaxis
SelectItem = select_item.SelectItem
select_item = select_item.select_item
separate = separate.separate
SplitAxis = split_axis.SplitAxis
split_axis = split_axis.split_axis
Squeeze = squeeze.Squeeze
squeeze = squeeze.squeeze
stack = stack.stack
Swapaxes = swapaxes.Swapaxes
swapaxes = swapaxes.swapaxes
Tile = tile.Tile
tile = tile.tile
Transpose = transpose.Transpose
transpose = transpose.transpose
TransposeSequence = transpose_sequence.TransposeSequence
transpose_sequence = transpose_sequence.transpose_sequence
Where = where.Where
where = where.where

bilinear = bilinear.bilinear
convolution_2d = convolution_2d.convolution_2d
convolution_nd = convolution_nd.convolution_nd
deconvolution_2d = deconvolution_2d.deconvolution_2d
deconvolution_nd = deconvolution_nd.deconvolution_nd
dilated_convolution_2d = dilated_convolution_2d.dilated_convolution_2d
embed_id = embed_id.embed_id
linear = linear.linear
NStepLSTM = n_step_lstm.NStepLSTM
n_step_lstm = n_step_lstm.n_step_lstm

Accuracy = accuracy.Accuracy
accuracy = accuracy.accuracy
BinaryAccuracy = binary_accuracy.BinaryAccuracy
binary_accuracy = binary_accuracy.binary_accuracy
ClassificationSummary = classification_summary_.ClassificationSummary
classification_summary = classification_summary_.classification_summary
precision = classification_summary_.precision
recall = classification_summary_.recall
f1_score = classification_summary_.f1_score

argmax_crf1d = crf1d.argmax_crf1d
bernoulli_nll = vae.bernoulli_nll
BinaryHierarchicalSoftmax = hierarchical_softmax.BinaryHierarchicalSoftmax
black_out = black_out.black_out
Contrastive = contrastive.Contrastive
contrastive = contrastive.contrastive
crf1d = crf1d.crf1d
CrossCovariance = cross_covariance.CrossCovariance
cross_covariance = cross_covariance.cross_covariance
gaussian_kl_divergence = vae.gaussian_kl_divergence
gaussian_nll = vae.gaussian_nll
Hinge = hinge.Hinge
hinge = hinge.hinge
HuberLoss = huber_loss.HuberLoss
huber_loss = huber_loss.huber_loss
MeanSquaredError = mean_squared_error.MeanSquaredError
mean_squared_error = mean_squared_error.mean_squared_error
negative_sampling = negative_sampling.negative_sampling
SigmoidCrossEntropy = sigmoid_cross_entropy.SigmoidCrossEntropy
sigmoid_cross_entropy = sigmoid_cross_entropy.sigmoid_cross_entropy
SoftmaxCrossEntropy = softmax_cross_entropy.SoftmaxCrossEntropy
softmax_cross_entropy = softmax_cross_entropy.softmax_cross_entropy
Triplet = triplet.Triplet
triplet = triplet.triplet
vstack = vstack.vstack

absolute = basic_math.absolute
ArgMax = minmax.ArgMax
argmax = minmax.argmax
ArgMin = minmax.ArgMin
argmin = minmax.argmin
Arccos = trigonometric.Arccos
arccos = trigonometric.arccos
Arcsin = trigonometric.Arcsin
arcsin = trigonometric.arcsin
Arctan = trigonometric.Arctan
arctan = trigonometric.arctan
BatchDet = det.BatchDet
batch_det = det.batch_det
BatchInv = inv.BatchInv
batch_inv = inv.batch_inv
BatchL2NormSquared = batch_l2_norm_squared.BatchL2NormSquared
batch_l2_norm_squared = batch_l2_norm_squared.batch_l2_norm_squared
BatchMatMul = matmul.BatchMatMul
batch_matmul = matmul.batch_matmul
bias = bias.bias
Ceil = ceil.Ceil
ceil = ceil.ceil
Clip = clip.Clip
clip = clip.clip
Cos = trigonometric.Cos
cos = trigonometric.cos
Cosh = hyperbolic.Cosh
cosh = hyperbolic.cosh
det = det.det
Exp = exponential.Exp
exp = exponential.exp
Expm1 = exponential_m1.Expm1
expm1 = exponential_m1.expm1
Floor = floor.Floor
floor = floor.floor
Identity = identity.Identity
identity = identity.identity
Inv = inv.Inv
inv = inv.inv
LinearInterpolate = linear_interpolate.LinearInterpolate
linear_interpolate = linear_interpolate.linear_interpolate
Log = exponential.Log
log = exponential.log
Log10 = exponential.Log10
log10 = exponential.log10
Log1p = logarithm_1p.Log1p
log1p = logarithm_1p.log1p
Log2 = exponential.Log2
log2 = exponential.log2
LogSumExp = logsumexp.LogSumExp
logsumexp = logsumexp.logsumexp
MatMul = matmul.MatMul
matmul = matmul.matmul
Max = minmax.Max
max = minmax.max
Maximum = maximum.Maximum
maximum = maximum.maximum
Minimum = minimum.Minimum
minimum = minimum.minimum
Min = minmax.Min
min = minmax.min
rsqrt = sqrt.rsqrt
scale = scale.scale
Sin = trigonometric.Sin
sin = trigonometric.sin
Sinh = hyperbolic.Sinh
sinh = hyperbolic.sinh
Sqrt = sqrt.Sqrt
sqrt = sqrt.sqrt
Square = square.Square
square = square.square
SquaredDifference = squared_difference.SquaredDifference
squared_difference = squared_difference.squared_difference
Sum = sum.Sum
sum = sum.sum
Tan = trigonometric.Tan
tan = trigonometric.tan

Dropout = dropout.Dropout
dropout = dropout.dropout
Gaussian = gaussian.Gaussian
gaussian = gaussian.gaussian
Zoneout = zoneout.Zoneout
zoneout = zoneout.zoneout

fixed_batch_normalization = batch_normalization.fixed_batch_normalization
batch_normalization = batch_normalization.batch_normalization
LocalResponseNormalization = \
    local_response_normalization.LocalResponseNormalization
local_response_normalization = \
    local_response_normalization.local_response_normalization
NormalizeL2 = l2_normalization.NormalizeL2
normalize = l2_normalization.normalize

AveragePooling2D = average_pooling_2d.AveragePooling2D
average_pooling_2d = average_pooling_2d.average_pooling_2d
MaxPooling2D = max_pooling_2d.MaxPooling2D
max_pooling_2d = max_pooling_2d.max_pooling_2d
ROIPooling2D = roi_pooling_2d.ROIPooling2D
roi_pooling_2d = roi_pooling_2d.roi_pooling_2d
SpatialPyramidPooling2D = spatial_pyramid_pooling_2d.SpatialPyramidPooling2D
spatial_pyramid_pooling_2d = \
    spatial_pyramid_pooling_2d.spatial_pyramid_pooling_2d

Unpooling2D = unpooling_2d.Unpooling2D
unpooling_2d = unpooling_2d.unpooling_2d

Forget = forget.Forget
forget = forget.forget

# Import for backward compatibility
PReLU = links_prelu.PReLU

Bilinear = links_bilinear.Bilinear
Convolution2D = links_convolution_2d.Convolution2D
DilatedConvolution2D = links_dilated_convolution_2d.DilatedConvolution2D
EmbedID = links_embed_id.EmbedID
Inception = inception.Inception
InceptionBN = inceptionbn.InceptionBN
Linear = links_linear.Linear
Parameter = parameter.Parameter

NegativeSampling = links_negative_sampling.NegativeSampling

BatchNormalization = links_batch_normalization.BatchNormalization

Space2Depth = space2depth.Space2Depth
space2depth = space2depth.space2depth



import contextlib

from cupy.cuda import compiler
from cupy.cuda import device
from cupy.cuda import function
from cupy.cuda import memory
from cupy.cuda import pinned_memory
from cupy.cuda import profiler
from cupy.cuda import stream

compile_with_cache = compiler.compile_with_cache

Device = device.Device
get_cublas_handle = device.get_cublas_handle
get_device_id = device.get_device_id

alloc = memory.alloc
Memory = memory.Memory
MemoryPointer = memory.MemoryPointer
MemoryPool = memory.MemoryPool
set_allocator = memory.set_allocator

alloc_pinned_memory = pinned_memory.alloc_pinned_memory
PinnedMemory = pinned_memory.PinnedMemory
PinnedMemoryPointer = pinned_memory.PinnedMemoryPointer
PinnedMemoryPool = pinned_memory.PinnedMemoryPool
set_pinned_memory_allocator = pinned_memory.set_pinned_memory_allocator

Function = function.Function
Module = function.Module

Event = stream.Event
Stream = stream.Stream
get_elapsed_time = stream.get_elapsed_time


@contextlib.contextmanager
def profile():
    """Enable CUDA profiling during with statement.

    This function enables profiling on entering a with statement, and disables
    profiling on leaving the statement.

    >>> with cupy.cuda.profile():
    ...    # do something you want to measure
    ...    pass

    """
    profiler.start()
    try:
        yield
    finally:
        profiler.stop()



"""Collection of :class:`~chainer.Link` implementations."""

from chainer.links.activation import maxout
from chainer.links.activation import prelu
from chainer.links.connection import bias
from chainer.links.connection import bilinear
from chainer.links.connection import convolution_2d
from chainer.links.connection import convolution_nd
from chainer.links.connection import deconvolution_2d
from chainer.links.connection import deconvolution_nd
from chainer.links.connection import dilated_convolution_2d
from chainer.links.connection import embed_id
from chainer.links.connection import gru
from chainer.links.connection import highway
from chainer.links.connection import inception
from chainer.links.connection import inceptionbn
from chainer.links.connection import linear
from chainer.links.connection import lstm
from chainer.links.connection import mlp_convolution_2d
from chainer.links.connection import n_step_lstm
from chainer.links.connection import parameter
from chainer.links.connection import peephole
from chainer.links.connection import scale
from chainer.links.connection import zoneoutlstm
from chainer.links.loss import black_out
from chainer.links.loss import crf1d
from chainer.links.loss import hierarchical_softmax
from chainer.links.loss import negative_sampling
from chainer.links.model import classifier
from chainer.links.normalization import batch_normalization


Maxout = maxout.Maxout
PReLU = prelu.PReLU

Bias = bias.Bias
Bilinear = bilinear.Bilinear
Convolution2D = convolution_2d.Convolution2D
ConvolutionND = convolution_nd.ConvolutionND
Deconvolution2D = deconvolution_2d.Deconvolution2D
DeconvolutionND = deconvolution_nd.DeconvolutionND
DilatedConvolution2D = dilated_convolution_2d.DilatedConvolution2D
EmbedID = embed_id.EmbedID
GRU = gru.GRU
StatefulGRU = gru.StatefulGRU
Highway = highway.Highway
Inception = inception.Inception
InceptionBN = inceptionbn.InceptionBN
Linear = linear.Linear
LSTM = lstm.LSTM
StatelessLSTM = lstm.StatelessLSTM
MLPConvolution2D = mlp_convolution_2d.MLPConvolution2D
NStepLSTM = n_step_lstm.NStepLSTM
Parameter = parameter.Parameter
StatefulPeepholeLSTM = peephole.StatefulPeepholeLSTM
Scale = scale.Scale
StatefulZoneoutLSTM = zoneoutlstm.StatefulZoneoutLSTM

BlackOut = black_out.BlackOut
CRF1d = crf1d.CRF1d
BinaryHierarchicalSoftmax = hierarchical_softmax.BinaryHierarchicalSoftmax
NegativeSampling = negative_sampling.NegativeSampling

Classifier = classifier.Classifier

BatchNormalization = batch_normalization.BatchNormalization



from chainer.training import extension
from chainer.training import trainer
from chainer.training import trigger
from chainer.training import updater


Extension = extension.Extension
make_extension = extension.make_extension
PRIORITY_WRITER = extension.PRIORITY_WRITER
PRIORITY_EDITOR = extension.PRIORITY_EDITOR
PRIORITY_READER = extension.PRIORITY_READER

Trainer = trainer.Trainer

IntervalTrigger = trigger.IntervalTrigger
get_trigger = trigger.get_trigger

Updater = updater.Updater
StandardUpdater = updater.StandardUpdater
ParallelUpdater = updater.ParallelUpdater



import numpy

from chainer import cuda
from chainer import function
from chainer.utils import type_check


class Triplet(function.Function):

    """Triplet loss function."""

    def __init__(self, margin):
        if margin <= 0:
            raise ValueError('margin should be positive value.')
        self.margin = margin

    def check_type_forward(self, in_types):
        type_check.expect(in_types.size() == 3)

        type_check.expect(
            in_types[0].dtype == numpy.float32,
            in_types[1].dtype == numpy.float32,
            in_types[2].dtype == numpy.float32,
            in_types[0].shape == in_types[1].shape,
            in_types[0].shape == in_types[2].shape,
            in_types[0].shape[0] > 0
        )

    def forward(self, inputs):
        xp = cuda.get_array_module(*inputs)

        anchor, positive, negative = inputs
        N = anchor.shape[0]

        dist = xp.sum(
            (anchor - positive) ** 2 - (anchor - negative) ** 2,
            axis=1) + self.margin
        self.dist_hinge = xp.maximum(dist, 0)
        loss = xp.sum(self.dist_hinge) / N

        return xp.array(loss, dtype=numpy.float32),

    def backward(self, inputs, gy):
        xp = cuda.get_array_module(*inputs)

        anchor, positive, negative = inputs
        N = anchor.shape[0]

        x_dim = anchor.shape[1]
        tmp = xp.repeat(self.dist_hinge[:, None], x_dim, axis=1)
        mask = xp.array(tmp > 0, dtype=numpy.float32)

        tmp = 2 * gy[0] * mask / N
        gx0 = (tmp * (negative - positive)).astype(numpy.float32)
        gx1 = (tmp * (positive - anchor)).astype(numpy.float32)
        gx2 = (tmp * (anchor - negative)).astype(numpy.float32)

        return gx0, gx1, gx2


def triplet(anchor, positive, negative, margin=0.2):
    """Computes triplet loss.

    It takes a triplet of variables as inputs, :math:`a`, :math:`p` and
    :math:`n`: anchor, positive example and negative example respectively.
    The triplet defines a relative similarity between samples.
    Let :math:`N` and :math:`K` denote mini-batch size and the dimension of
    input variables, respectively. The shape of all input variables should be
    :math:`(N, K)`.

    .. math::
        L(a, p, n) = \\frac{1}{N} \\left( \\sum_{i=1}^N \\max \{d(a_i, p_i)
            - d(a_i, n_i) + {\\rm margin}, 0\} \\right)

    where :math:`d(x_i, y_i) = \\| {\\bf x}_i - {\\bf y}_i \\|_2^2`.

    Args:
        anchor (~chainer.Variable): The anchor example variable. The shape
            should be :math:`(N, K)`, where :math:`N` denotes the minibatch
            size, and :math:`K` denotes the dimension of the anchor.
        positive (~chainer.Variable): The positive example variable. The shape
            should be the same as anchor.
        negative (~chainer.Variable): The negative example variable. The shape
            should be the same as anchor.
        margin (float): A parameter for triplet loss. It should be a positive
            value.

    Returns:
        ~chainer.Variable: A variable holding a scalar that is the loss value
            calculated by the above equation.

    .. note::
        This cost can be used to train triplet networks. See `Learning \
        Fine-grained Image Similarity with Deep Ranking \
        <https://arxiv.org/abs/1404.4661>`_ for details.
    """
    return Triplet(margin)(anchor, positive, negative)



import numpy
import six

from chainer import cuda
from chainer import function
from chainer.utils import type_check


class NegativeSamplingFunction(function.Function):

    ignore_label = -1

    def __init__(self, sampler, sample_size):
        self.sampler = sampler
        self.sample_size = sample_size

    def _make_samples(self, t):
        if hasattr(self, 'samples'):
            return self.samples  # for testing

        size = int(t.shape[0])
        # first one is the positive, and others are sampled negatives
        samples = self.sampler((size, self.sample_size + 1))
        samples[:, 0] = t
        self.samples = samples

    def check_type_forward(self, in_types):
        type_check.expect(in_types.size() == 3)
        x_type, t_type, w_type = in_types

        type_check.expect(
            x_type.dtype == numpy.float32,
            x_type.ndim == 2,
            t_type.dtype == numpy.int32,
            t_type.ndim == 1,
            x_type.shape[0] == t_type.shape[0],
            w_type.dtype == numpy.float32,
            w_type.ndim == 2,
        )

    def forward_cpu(self, inputs):
        x, t, W = inputs
        self.ignore_mask = (t != self.ignore_label)
        self._make_samples(t)

        loss = numpy.float32(0.0)
        for ix, k in six.moves.zip(x[self.ignore_mask],
                                   self.samples[self.ignore_mask]):
            w = W[k]
            f = w.dot(ix)
            f[0] *= -1  # positive sample
            loss += numpy.sum(numpy.logaddexp(f, 0))
        return numpy.array(loss, numpy.float32),

    def forward_gpu(self, inputs):
        x, t, W = inputs
        self.ignore_mask = (t != self.ignore_label)
        n_in = x.shape[1]
        self._make_samples(t)

        self.wx = cuda.elementwise(
            'raw T W, raw T x, bool mask, S k, int32 c, int32 m', 'T wx',
            '''
            T f = 0;
            if (mask == 1) {
                for (int j = 0; j < c; ++j) {
                  int x_ind[] = {(i / m), j};
                  int w_ind[] = {k, j};
                  f += x[x_ind] * W[w_ind];
                }
            }
            wx = f;
            ''',
            'negative_sampling_wx'
        )(W, x, self.ignore_mask[:, None], self.samples, n_in,
          self.sample_size + 1)

        y = cuda.elementwise(
            'T wx, int32 c, int32 m', 'T y',
            '''
            T f = wx;
            if (i % m == 0) {
              f = -f;
            }
            T loss;
            if (f < 0) {
              loss = __logf(1 + __expf(f));
            } else {
              loss = f + __logf(1 + __expf(-f));
            }
            y = loss;
            ''',
            'negative_sampling_forward'
        )(self.wx, n_in, self.sample_size + 1)
        # TODO(okuta): merge elementwise
        loss = cuda.cupy.sum(y * self.ignore_mask[:, None].astype('float32'))
        return loss,

    def backward_cpu(self, inputs, grads):
        x, t, W = inputs
        gloss, = grads

        gx = numpy.zeros_like(x)
        gW = numpy.zeros_like(W)
        for i, (ix, k) in enumerate(six.moves.zip(
                x[self.ignore_mask], self.samples[self.ignore_mask])):
            w = W[k]
            f = w.dot(ix)

            # g == -y * gloss / (1 + exp(yf))
            f[0] *= -1
            g = gloss / (1 + numpy.exp(-f))
            g[0] *= -1

            gx[i] = g.dot(w)
            for ik, ig in six.moves.zip(k, g):
                gW[ik] += ig * ix
        return gx, None, gW

    def backward_gpu(self, inputs, grads):
        cupy = cuda.cupy
        x, t, W = inputs
        gloss, = grads

        n_in = x.shape[1]
        g = cuda.elementwise(
            'T wx, raw T gloss, int32 m', 'T g',
            '''
            T y;
            if (i % m == 0) {
              y = 1;
            } else {
              y = -1;
            }

            g = -y * gloss[0] / (1.0f + __expf(wx * y));
            ''',
            'negative_sampling_calculate_g'
        )(self.wx, gloss, self.sample_size + 1)
        gx = cupy.zeros_like(x)
        cuda.elementwise(
            'raw T g, raw T W, bool mask, raw S k, int32 c, int32 m', 'T gx',
            '''
            int d = i / c;
            T w = 0;
            if (mask == 1){
                for (int j = 0; j < m; ++j) {
                  w += g[d * m + j] * W[k[d * m + j] * c + i % c];
                }
            }
            gx = w;
            ''',
            'negative_sampling_calculate_gx'
        )(g, W, self.ignore_mask[:, None], self.samples, n_in,
          self.sample_size + 1, gx)
        gW = cupy.zeros_like(W)
        cuda.elementwise(
            'T g, raw T x, S k, bool mask, int32 c, int32 m',
            'raw T gW',
            '''
            T gi = g;
            if (mask == 1) {
                for (int j = 0; j < c; ++j) {
                  atomicAdd(&gW[k * c + j], gi * x[(i / m) * c + j]);
                }
            }
            ''',
            'negative_sampling_calculate_gw'
        )(g, x, self.samples, self.ignore_mask[:, None], n_in,
          self.sample_size + 1, gW)
        return gx, None, gW


def negative_sampling(x, t, W, sampler, sample_size):
    """Negative sampling loss function.

    In natural language processing, especially language modeling, the number of
    words in a vocabulary can be very large.
    Therefore, you need to spend a lot of time calculating the gradient of the
    embedding matrix.

    By using the negative sampling trick you only need to calculate the
    gradient for a few sampled negative examples.

    The objective function is below:

    .. math::

       f(x, p) = \\log \\sigma(x^\\top w_p) +  \\
       k E_{i \\sim P(i)}[\\log \\sigma(- x^\\top w_i)],

    where :math:`\\sigma(\\cdot)` is a sigmoid function, :math:`w_i` is the
    weight vector for the word :math:`i`, and :math:`p` is a positive example.
    It is approximated with :math:`k` examples :math:`N` sampled from
    probability :math:`P(i)`, like this:

    .. math::

       f(x, p) \\approx \\log \\sigma(x^\\top w_p) +  \\
       \\sum_{n \\in N} \\log \\sigma(-x^\\top w_n).

    Each sample of :math:`N` is drawn from the word distribution :math:`P(w)`.
    This is calculated as :math:`P(w) = \\frac{1}{Z} c(w)^\\alpha`, where
    :math:`c(w)` is the unigram count of the word :math:`w`, :math:`\\alpha` is
    a hyper-parameter, and :math:`Z` is the normalization constant.

    Args:
        x (~chainer.Variable): Batch of input vectors.
        t (~chainer.Variable): Vector of ground truth labels.
        W (~chainer.Variable): Weight matrix.
        sampler (~types.FunctionType): Sampling function. It takes a shape and
            returns an integer array of the shape. Each element of this array
            is a sample from the word distribution.
            A :class:`~chainer.utils.WalkerAlias` object built with the power
            distribution of word frequency is recommended.
        sample_size (int): Number of samples.

    See: `Distributed Representations of Words and Phrases and their\
         Compositionality <http://arxiv.org/abs/1310.4546>`_

    .. seealso:: :class:`~chainer.links.NegativeSampling`.

    """
    return NegativeSamplingFunction(sampler, sample_size)(x, t, W)



import collections

import numpy

from chainer import cuda
from chainer import function
from chainer.utils import conv
from chainer.utils import type_check


if cuda.cudnn_enabled:
    cudnn = cuda.cudnn
    libcudnn = cudnn.cudnn
    _cudnn_version = libcudnn.getVersion()


def _check_cudnn_acceptable_type(x_dtype):
    return _cudnn_version >= 3000 or x_dtype != numpy.float16


def _pair(x):
    if isinstance(x, collections.Iterable):
        return x
    return x, x


class Pooling2D(function.Function):

    """Base class of pooling function over a set of 2d planes."""

    def __init__(self, ksize, stride=None, pad=0, cover_all=True,
                 use_cudnn=True):
        if stride is None:
            stride = ksize

        self.kh, self.kw = _pair(ksize)
        self.sy, self.sx = _pair(stride)
        self.ph, self.pw = _pair(pad)

        self.cover_all = cover_all
        self.use_cudnn = use_cudnn

    def check_type_forward(self, in_types):
        type_check.expect(
            in_types.size() == 1,
            in_types[0].dtype.kind == 'f',
            in_types[0].ndim == 4
        )

    def forward_gpu(self, x):
        # Implementation using cudnn
        x = x[0]
        n, c, h, w = x.shape
        y_h = conv.get_conv_outsize(
            h, self.kh, self.sy, self.ph, self.cover_all)
        y_w = conv.get_conv_outsize(
            w, self.kw, self.sx, self.pw, self.cover_all)
        y = cuda.cupy.empty((n, c, y_h, y_w), dtype=x.dtype)

        handle = cudnn.get_handle()
        pool_desc = self.create_pool_desc()
        x_desc = cudnn.create_tensor_descriptor(x)
        y_desc = cudnn.create_tensor_descriptor(y)

        oz_dtype = 'd' if x.dtype == 'd' else 'f'
        one = numpy.array(1, dtype=oz_dtype).ctypes
        zero = numpy.array(0, dtype=oz_dtype).ctypes
        libcudnn.poolingForward(
            handle, pool_desc.value, one.data, x_desc.value,
            x.data.ptr, zero.data, y_desc.value, y.data.ptr)
        self.y = y

        return y,

    def backward_gpu(self, x, gy):
        # Implementation using cudnn
        x = x[0]
        handle = cudnn.get_handle()
        pool_desc = self.create_pool_desc()

        # Pooling of cuDNNv2 does not seem to support non-contiguous gradients
        gy = cuda.cupy.ascontiguousarray(gy[0])

        x_desc = cudnn.create_tensor_descriptor(x)
        y_desc = cudnn.create_tensor_descriptor(gy)

        oz_dtype = 'd' if x.dtype == 'd' else 'f'
        one = numpy.array(1, dtype=oz_dtype).ctypes
        zero = numpy.array(0, dtype=oz_dtype).ctypes
        gx = cuda.cupy.empty_like(x)
        libcudnn.poolingBackward(
            handle, pool_desc.value, one.data, y_desc.value,
            self.y.data.ptr, y_desc.value, gy.data.ptr, x_desc.value,
            x.data.ptr, zero.data, x_desc.value, gx.data.ptr)
        return gx,

    def create_pool_desc(self):
        raise NotImplementedError()



from chainer.datasets import cifar
from chainer.datasets import dict_dataset
from chainer.datasets import image_dataset
from chainer.datasets import mnist
from chainer.datasets import ptb
from chainer.datasets import sub_dataset
from chainer.datasets import tuple_dataset


DictDataset = dict_dataset.DictDataset
ImageDataset = image_dataset.ImageDataset
LabeledImageDataset = image_dataset.LabeledImageDataset
SubDataset = sub_dataset.SubDataset
TupleDataset = tuple_dataset.TupleDataset

get_cross_validation_datasets = sub_dataset.get_cross_validation_datasets
get_cross_validation_datasets_random = (
    sub_dataset.get_cross_validation_datasets_random)

split_dataset = sub_dataset.split_dataset
split_dataset_random = sub_dataset.split_dataset_random


# examples

get_cifar10 = cifar.get_cifar10
get_cifar100 = cifar.get_cifar100
get_mnist = mnist.get_mnist
get_ptb_words = ptb.get_ptb_words
get_ptb_words_vocabulary = ptb.get_ptb_words_vocabulary



from chainer.training.extensions import _snapshot
from chainer.training.extensions import computational_graph
from chainer.training.extensions import evaluator
from chainer.training.extensions import exponential_shift
from chainer.training.extensions import linear_shift
from chainer.training.extensions import log_report
from chainer.training.extensions import print_report
from chainer.training.extensions import progress_bar
from chainer.training.extensions import value_observation


dump_graph = computational_graph.dump_graph
Evaluator = evaluator.Evaluator
ExponentialShift = exponential_shift.ExponentialShift
LinearShift = linear_shift.LinearShift
LogReport = log_report.LogReport
snapshot = _snapshot.snapshot
snapshot_object = _snapshot.snapshot_object
PrintReport = print_report.PrintReport
ProgressBar = progress_bar.ProgressBar
observe_value = value_observation.observe_value
observe_lr = value_observation.observe_lr



from chainer import cuda
from chainer import function
from chainer.utils import type_check


class CReLU(function.Function):

    """Concatenated Rectified Linear Unit."""

    def __init__(self, axis=1):
        if not isinstance(axis, int):
            raise TypeError('axis must be an integer value')
        self.axis = axis

    def check_type_forward(self, in_types):
        type_check.expect(
            in_types.size() == 1,
            in_types[0].dtype.kind == 'f',
            in_types[0].ndim > self.axis,
            in_types[0].ndim >= -self.axis
        )

    def get_output_shape(self, input_shape):
        output_shape = list(input_shape)
        output_shape[self.axis] *= 2
        return tuple(output_shape)

    def forward(self, x):
        x, = x
        xp = cuda.get_array_module(x)
        y = xp.empty(self.get_output_shape(x.shape), dtype=x.dtype)
        y_former, y_latter = xp.split(y, 2, axis=self.axis)
        zero = x.dtype.type(0)
        xp.maximum(zero, x, out=y_former)
        xp.maximum(zero, -x, out=y_latter)
        return y,

    def backward(self, x, gy):
        x, = x
        xp = cuda.get_array_module(x)
        gy, = gy
        gy_former, gy_latter = xp.split(gy, 2, axis=self.axis)
        return gy_former * (x > 0) - gy_latter * (-x > 0),


def crelu(x, axis=1):
    """Concatenated Rectified Linear Unit function.

    This function is expressed as :math:`f(x) = (\\max(0, x), \\max(0, -x))`,
    where two output values are concatenated along an axis.

    See: http://arxiv.org/abs/1603.05201

    Args:
        x (~chainer.Variable): Input variable.
        axis (int): Axis that the output values are concatenated along

    Returns:
        ~chainer.Variable: Output variable.

    """
    return CReLU(axis=axis)(x)
